#!/usr/bin/env python

import rclpy
from rclpy.node import Node
from geometry_msgs.msg import Twist
from std_msgs.msg import Empty, Float32, Bool
from sensor_msgs.msg import Joy

# 手柄按键和轴的映射
BUTTON_RT = 5                # RT 按键的映射
AXIS_LEFT_HORIZONTAL = 0      # 左摇杆水平方向（左右）
AXIS_LEFT_VERTICAL = 1        # 左摇杆垂直方向（上下）
AXIS_RIGHT_HORIZONTAL = 3     # 右摇杆水平方向（左右）
AXIS_RIGHT_VERTICAL = 4       # 右摇杆垂直方向（上下）

class JoyTeleop(Node):
    def __init__(self):
        super().__init__('joy_teleop')

        # 创建发布器
        self.linear_x_pub = self.create_publisher(Float32, "/linear_x", 10)
        self.linear_y_pub = self.create_publisher(Float32, "/linear_y", 10)
        self.linear_z_pub = self.create_publisher(Float32, "/linear_z", 10)
        self.angular_z_pub = self.create_publisher(Float32, "/angular_z", 10)
        self.hover_pub = self.create_publisher(Bool, "/hover", 10)
        self.takeoff_pub = self.create_publisher(Empty, '/bebop/takeoff', 10)
        self.land_pub = self.create_publisher(Empty, '/bebop/land', 10)

        # 创建订阅器
        self.joy_sub = self.create_subscription(Joy, '/joy', self.joy_callback, 10)

        # 初始化状态
        self.is_flying = False
        self.axis_tolerance = 0.75  # 按键按下的检测阈值

    def joy_callback(self, msg):
        # 检查 RT 按键的状态
        if msg.axes[BUTTON_RT] < -1.0 + self.axis_tolerance:
            if not self.is_flying:
                # 起飞控制
                self.takeoff_pub.publish(Empty())
                self.is_flying = True
                self.get_logger().info("Takeoff triggered")
            else:
                # RT 按键保持按下，发布速度命令
                hover_msg = Bool()
                hover_msg.data = False  # 不处于悬停模式
                self.hover_pub.publish(hover_msg)

                # 发布线速度和角速度
                linear_x = Float32()
                linear_y = Float32()
                linear_z = Float32()
                angular_z = Float32()

                # 设置线速度和角速度值
                linear_x.data = msg.axes[AXIS_LEFT_VERTICAL] * 0.5
                linear_y.data = msg.axes[AXIS_LEFT_HORIZONTAL] * 0.5
                linear_z.data = msg.axes[AXIS_RIGHT_VERTICAL] * 0.5
                angular_z.data = msg.axes[AXIS_RIGHT_HORIZONTAL] * 0.5

                # 发布速度命令
                self.linear_x_pub.publish(linear_x)
                self.linear_y_pub.publish(linear_y)
                self.linear_z_pub.publish(linear_z)
                self.angular_z_pub.publish(angular_z)

        else:
            # 如果 RT 被释放，触发降落
            if self.is_flying:
                self.land_pub.publish(Empty())
                self.is_flying = False
                self.get_logger().info("Landing triggered")

def main(args=None):
    rclpy.init(args=args)
    joy_teleop = JoyTeleop()
    rclpy.spin(joy_teleop)
    joy_teleop.destroy_node()
    rclpy.shutdown()

if __name__ == '__main__':
    main()
