# !/usr/bin/env python

import rclpy
from rclpy.node import Node
from geometry_msgs.msg import Twist
from std_msgs.msg import Float32, Bool
from nav_msgs.msg import Odometry
from sensor_msgs.msg import Joy
from threading import Lock
from .PID import PID
import datetime

# Define button mappings for the remote controller
AXIS_CROSS_HORIZONTAL = 6  # D-pad horizontal direction
AXIS_CROSS_VERTICAL = 7    # D-pad vertical direction

class SpeedController(Node):
    def __init__(self):
        super().__init__('speed_controller')

        # Initialize attributes
        self.target_twist = Twist()  # Target velocity
        self.lock = Lock()           # Mutex lock to protect target velocity
        self.hover_mode = False      # Hover mode
        self.odom_velocity = Twist()

        # Initialize PID controller gains based on Ziegler-Nichols method
        Kp_initial = 1.5
        self.gains = {
            'Kp': Kp_initial,
            'Ki': Kp_initial / 2,       # Ki = Kp / 2 = 0.75
            'Kd': Kp_initial / 8        # Kd = Kp / 8 = 0.1875
        }
        self.pid_controller_x = PID(self.gains)
        self.pid_controller_y = PID(self.gains)

        # Create publishers
        self.cmd_vel_pub = self.create_publisher(Twist, '/bebop/cmd_vel', 10)
        self.target_vel_pub = self.create_publisher(Twist, '/target_vel', 10)

        # Monitoring topic publishers
        self.linear_x_pub = self.create_publisher(Float32, '/cmd_vel_mux/input/teleop/linear/x', 10)
        self.linear_y_pub = self.create_publisher(Float32, '/cmd_vel_mux/input/teleop/linear/y', 10)
        self.linear_z_pub = self.create_publisher(Float32, '/cmd_vel_mux/input/teleop/linear/z', 10)
        self.angular_z_pub = self.create_publisher(Float32, '/cmd_vel_mux/input/teleop/angular/z', 10)
        
        # Create subscribers
        self.odom_sub = self.create_subscription(Odometry, '/odom', self.odom_callback, 10)
        self.linear_x_sub = self.create_subscription(Float32, '/linear_x', self.linear_x_callback, 10)
        self.linear_y_sub = self.create_subscription(Float32, '/linear_y', self.linear_y_callback, 10)
        self.linear_z_sub = self.create_subscription(Float32, '/linear_z', self.linear_z_callback, 10)
        self.angular_z_sub = self.create_subscription(Float32, '/angular_z', self.angular_z_callback, 10)
        self.hover_sub = self.create_subscription(Bool, '/hover', self.hover_callback, 10)
        self.joy_sub = self.create_subscription(Joy, '/joy', self.joy_callback, 10)

        # Control loop
        self.timer = self.create_timer(0.1, self.control_loop)  # Control frequency 10Hz
        self.t0 = self.get_clock().now()

    def reset_target_twist(self):
        # Reset target velocity
        with self.lock:
            self.target_twist = Twist()

    def odom_callback(self, msg):
        # Handle /odom topic callback
        with self.lock:
            self.odom_velocity = msg.twist.twist

    def linear_x_callback(self, msg):
        with self.lock:
            self.target_twist.linear.x = msg.data
        self.hover_mode = False  # Exit hover mode

    def linear_y_callback(self, msg):
        with self.lock:
            self.target_twist.linear.y = msg.data
        self.hover_mode = False  # Exit hover mode

    def linear_z_callback(self, msg):
        with self.lock:
            self.target_twist.linear.z = msg.data
        self.hover_mode = False  # Exit hover mode

    def angular_z_callback(self, msg):
        with self.lock:
            self.target_twist.angular.z = msg.data
        self.hover_mode = False  # Exit hover mode

    def hover_callback(self, msg):
        self.hover_mode = msg.data
        if self.hover_mode:
            # Reset target velocity and PID errors
            self.reset_target_twist()
            self.send_null_twist()  # Immediately send a null Twist message

    def send_null_twist(self):
        # Publish a null Twist message to trigger hover
        null_twist = Twist()
        self.cmd_vel_pub.publish(null_twist)
        self.target_vel_pub.publish(null_twist)

    def rclpy_time_to_datetime(self, rclpy_time):
        """Convert rclpy time to Python datetime object"""
        seconds = rclpy_time.nanoseconds * 1e-9
        return datetime.datetime.fromtimestamp(seconds)

    def joy_callback(self, msg):
        """Adjust Kp value using D-pad"""
        if msg.axes[AXIS_CROSS_VERTICAL] > 0.5:
            self.gains['Kp'] += 0.1
        elif msg.axes[AXIS_CROSS_VERTICAL] < -0.5:
            self.gains['Kp'] -= 0.1

        # Update gains for PID controllers
        self.pid_controller_x.gains = self.gains.copy()
        self.pid_controller_y.gains = self.gains.copy()

    def control_loop(self):
        if self.hover_mode:
            # In hover mode, continuously send a null Twist message
            self.send_null_twist()
        else:
            # In normal mode, publish target velocity to cmd_vel and target_vel
            with self.lock:
                cmd_vel_msg = Twist()
                cmd_vel_msg.linear.x = max(min(self.target_twist.linear.x, 1.0), -1.0)
                cmd_vel_msg.linear.y = max(min(self.target_twist.linear.y, 1.0), -1.0)
                cmd_vel_msg.linear.z = max(min(self.target_twist.linear.z, 1.0), -1.0)
                cmd_vel_msg.angular.z = max(min(self.target_twist.angular.z, 1.0), -1.0)

                # Update target values for the PID controller
                self.pid_controller_x.update(self.rclpy_time_to_datetime(self.get_clock().now()), cmd_vel_msg.linear.x - self.odom_velocity.linear.x)
                self.pid_controller_y.update(self.rclpy_time_to_datetime(self.get_clock().now()), cmd_vel_msg.linear.y - self.odom_velocity.linear.y)

                # Calculate and set adjusted commands
                cmd_vel_msg.linear.x = self.pid_controller_x.command
                cmd_vel_msg.linear.y = self.pid_controller_y.command

                # Publish adjusted velocity command
                self.cmd_vel_pub.publish(cmd_vel_msg)
                self.target_vel_pub.publish(self.target_twist)

                # Publish each directional velocity to new topics for monitoring in rqt_plot
                self.linear_x_pub.publish(Float32(data=self.target_twist.linear.x))
                self.linear_y_pub.publish(Float32(data=self.target_twist.linear.y))
                self.linear_z_pub.publish(Float32(data=self.target_twist.linear.z))
                self.angular_z_pub.publish(Float32(data=self.target_twist.angular.z))

def main(args=None):
    rclpy.init(args=args)

    speed_controller = SpeedController()
    rclpy.spin(speed_controller)

    speed_controller.destroy_node()
    rclpy.shutdown()


if __name__ == "__main__":
    main()
