import numpy as np
import cv2
import sys
from sklearn.cluster import KMeans
from scipy.cluster.hierarchy import linkage, fcluster
import math

def from_lsd(lines_lsd):
   """
   lines_lsd : the weird output of cv line segment detector.
   returns : [[x1, y1, x2, y2, a, b, c, L],
              [...                       ],
              ...
             ]

             where 
               - [(x1, y1), (x2, y2)] is the segment (in image coordinates).
               - ax + by + c = 0 is the line equation
               - (a, b) is the normal vector of the line
               - |(a, b)| = 1, b >= 0
               - L is the length of the segment.
   """
   if lines_lsd is None :
       return np.array([]).reshape((0,8))

   nlines = lines_lsd.shape[0]
   segments = None

   if nlines == 0 :
       return np.array([]).reshape((0,8))

   if nlines == 1 :
       segments = lines_lsd.squeeze().reshape((1,4))
   else :
       segments = lines_lsd.squeeze()

   x1 = segments[:, 0]
   y1 = segments[:, 1]
   x2 = segments[:, 2]
   y2 = segments[:, 3]
   b = x1 - x2
   a = y2 - y1
   c = -(a * x1 + b * y1)

   # Ensure that the normal vectors are always
   # pointing toward positive y
   where_b_is_neg = b < 0
   c[where_b_is_neg] *= -1.0
   b[where_b_is_neg] *= -1.0
   a[where_b_is_neg] *= -1.0

   # The length of the segment
   norm = np.sqrt(a**2 + b**2)
   return np.stack([x1, y1, x2, y2, a/norm, b/norm, c/norm, norm], axis=1)


def draw_segments(img, segments, color, thickness):
   '''
       img : cv2 image
       segments : np.array(num_lines, 8)
       color  : 3 element tuple
       thickness : int
   '''
   for s in segments:
       cv2.line(img, (s[0], s[1]), (s[2], s[3]), color, thickness)


def draw_lines(img, segments, color, thickness):
   '''
       img : cv2 image
       segments : np.array(num_lines, 8)
       color  : 3 element tuple
       thickness : int
   '''

   for s in segments:
       a, b, c = s[4:7]
       # If the line is horizontal
       if a == 0:
           cv2.line(img, (0, -int(c/b)), (img.shape[1]-1, -int((c+a*(img.shape[1]-1))/b)), color, thickness)
       # If the line is vertical
       elif b == 0:
           cv2.line(img, (-int(c/a), 0), (-int(c/a), img.shape[0] - 1), color, thickness)
       else:
           # We look for the intersection point with the horizontal axis y=0 and y=height-1
           y = 0
           x = -c/a   #-(c - b*y)/a
           if x < 0:
               x = 0
               y = -(c + a * x) / b
           elif x >= img.shape[1]:
               x = img.shape[1] - 1
               y = -(c + a * x) / b
           first_point = (int(x), int(y))

           y = img.shape[0] - 1
           x = -(c+b*y)/a
           if x < 0:
               x = 0
               y = -(c + a * x) / b
           elif x >= img.shape[1]:
               x = img.shape[1] - 1
               y = -(c + a * x) / b
           second_point = (int(x), int(y))

           cv2.line(img, first_point, second_point, color, thickness)

def intersections(segments):
   intersections = []
   for i, si in enumerate(segments):
       for sj in segments[i+1:]:
           cross_product = np.cross(si[4:6], sj[4:6]) # [a1,b1] ^ [a2, b2]
           if cross_product != 0:
               coeff = 1.0 / cross_product

               intersections.append([coeff * np.cross(si[5:7]   , sj[5:7]), # [b1, c1] ^ [b2, c2]
                                     coeff * np.cross(sj[[4, 6]], si[[4, 6]])]) # -[a1, c1] ^ [a2, c2]
   return np.array(intersections)


def hierarchical_clustering(normal_vectors):
   Z = linkage(normal_vectors, method='ward')  # 'ward' minimizes variance, use other methods if needed
   return fcluster(Z, 2, criterion='maxclust')

def vanishing_lines(lines, normal_vectors):
   # Perform hierarchical/agglomerative clustering
   # Extract two largest clusters
   labels = hierarchical_clustering(normal_vectors)
   lines = np.concatenate((lines[labels == 1], lines[labels == 2]),axis=0)
   normal_vectors = np.concatenate((normal_vectors[labels == 1], normal_vectors[labels == 2]),axis=0)

   kmeans = KMeans(n_clusters=2, random_state=0).fit(normal_vectors)
   labels = kmeans.labels_
   # Séparer les lignes en fonction des clusters
   return lines[labels == 0], lines[labels == 1]

def horizontal_misplacement(frame, vanishing_point):

   ans=vanishing_point[0]/frame.shape[1]-0.5
   if ans<-0.15:
       return -0.25
   if ans>0.15:
       return 0.25
   return 0

def angle_indicator(line1, line2): #Ratio between left and right angles
   agl1 = np.pi/2 - np.arccos(abs(line1[0]))
   agl2 = np.pi/2 - np.arccos(abs(line2[0]))
   # Return the ratio of the two angles
#    return agl1/agl2
   agl=agl1/agl2
   if agl<0.7:
       return -0.5
   if agl>1.5:
       return 0.5
   return 0


def draw_function(
    img,
    Y,                   # The values, len(Y) = img.shape[1]
    hmin, hmax,          # The curve is plotted between image lines hmin and hmax.
    ymin, ymax,          # The value range [ymin, ymax] is mapped into [hmin, hmax].
    color, thickness):   # The color and thickness of the curve
    """
    Draws a curve over an image.
    
    img: np.array, the input image
    Y: np.array, the curve values (1D array)
    hmin, hmax: int, the vertical range for plotting the curve
    ymin, ymax: float, the value range for mapping the curve
    color: tuple, the color of the curve (e.g., (0, 255, 0) for green)
    thickness: int, the thickness of the curve line
    """
    # Normalize the curve values Y to fit within the vertical range [hmin, hmax]
    for i in range(len(Y) - 1):
        y1 = int(hmin + (hmax - hmin) * (Y[i] - ymin) / (ymax - ymin))
        y2 = int(hmin + (hmax - hmin) * (Y[i + 1] - ymin) / (ymax - ymin))
        cv2.line(img, (i, y1), (i + 1, y2), color, thickness)


import numpy as np
from scipy.signal import medfilt

def local_shift_computation(curve1, curve2, max_shift=20, window_radius=5):
    """
    优化局部位移计算算法
    """
    length = min(len(curve1), len(curve2))
    curve1 = curve1[:length]
    curve2 = curve2[:length]

    # 初始化误差矩阵
    error_matrix = np.full((length, max_shift + 1), float('inf'))

    for s in range(0, max_shift + 1):
        shifted_curve2 = np.roll(curve2, s)
        local_diff = (curve1 - shifted_curve2) ** 2
        error_matrix[:, s] = np.convolve(local_diff, np.ones((2 * window_radius + 1,)), mode='same')

    best_shifts = np.argmin(error_matrix, axis=1)
    smoothed_shifts = medfilt(best_shifts, kernel_size=21)

    return smoothed_shifts


# Test and plot the shift on an image
def plot_shift_on_image(img, shifts, color=(255, 0, 0), thickness=1):
    """
    Draws the shift curve on an image.
    
    img: np.array, the input image
    shifts: np.array, the shift values to plot
    color: tuple, the color of the curve
    thickness: int, the thickness of the curve line
    """
    hmin, hmax = 50, img.shape[0] - 50  # Define vertical range for plotting
    ymin, ymax = np.min(shifts), np.max(shifts)
    
    draw_function(img, shifts, hmin, hmax, ymin, ymax, color, thickness)
