import rclpy
from rclpy.node import Node
from sensor_msgs.msg import CompressedImage
import cv2
import numpy as np
from cv_bridge import CvBridge
from vanishing_point.my_line_library import local_shift_computation, draw_function

class OpticalFlowNode(Node):
    def __init__(self):
        super().__init__('optical_flow_node')
        self.subscription = self.create_subscription(
            CompressedImage,
            '/bebop/image_raw/compressed',
            self.listener_callback,
            10)
        self.bridge = CvBridge()

    def listener_callback(self, msg):
        np_arr = np.frombuffer(msg.data, np.uint8)
        image = cv2.imdecode(np_arr, cv2.IMREAD_COLOR)
        gray_image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)

        if not hasattr(self, 'previous_gray_image'):
            self.previous_gray_image = gray_image
            return

        row_index = 100
        curve1 = self.previous_gray_image[row_index, :]
        curve2 = gray_image[row_index, :]

        shifts = local_shift_computation(curve1, curve2)

        draw_function(image, curve1, hmin=0, hmax=image.shape[0], ymin=0, ymax=255, color=(255, 0, 0), thickness=1)
        draw_function(image, curve2, hmin=0, hmax=image.shape[0], ymin=0, ymax=255, color=(0, 165, 255), thickness=1)

        cv2.imshow("Optical Flow Image", image)
        print("Press Enter to view the next frame...")
        cv2.waitKey(0)  # Wait for Enter key

        self.previous_gray_image = gray_image

def main(args=None):
    rclpy.init(args=args)
    optical_flow_node = OpticalFlowNode()
    try:
        rclpy.spin(optical_flow_node)
    except KeyboardInterrupt:
        optical_flow_node.get_logger().info('Shutting down optical flow node...')
    finally:
        optical_flow_node.destroy_node()
        rclpy.shutdown()
        cv2.destroyAllWindows()

if __name__ == '__main__':
    main()
